"""
basicMLpy 1.0.2
=================


Provides
    1. Simple and effective machine learning algorithms for classification and regression
    2. Easy-to-use functions that allow the user to expand on the package's capabilities
    3. A Cross-Validation algorithm for the functions presented by the package
=================

Modules
    1.basicMLpy.regression
    2.basicMLpy.classification
    3.basicMLpy.nearest_neighbors
    4.basicMLpy.ensemble
    5.basicMLpy.model_selection
    6.basicMLpy.loss_functions
    7.basicMLpy.utils
=================


For more information on the package, please visit my Github page or the package page at pypi
Thanks for using the package and feel free to contact me for suggestions! 

"""